import os

from evidence.model.CEAction import CEAction

class CharacterizeEvidence():
    def __init__(self, occurence) -> None:
        self._actions = []
        self._occurence = occurence

    def _initSet(self, indicator, list):
        set_ = self._getSetForIndicator(indicator)
        for l in list:
            line = l.split("\t")
            if(indicator == "noise"):
                set_.add(line[0] + "\t" + line[1] + "\n")
            if(int(line[2]) >= self._occurence):                       # Only if evidence occures min 2 times
                set_.add(line[0] + "\t" + line[1] + "\n")
        
    def _getSetForIndicator(self, indicator):
        result = None
        for action in self._actions:
            if(indicator == action.name):
                result = action.set
        return result

    def _getCESetForIndicator(self, indicator):
        result = None
        for action in self._actions:
            if(indicator == action.name):
                result = action.ceSet
        return result
    
    def _writeToFile(self, indicator, path):
        w = []
        file = open(path, "w", encoding="utf-8")

        for line in self._getCESetForIndicator(indicator):
            w.append(line)
             
        for line in self._sortAlphabetically(w):
            file.write(line)

        print("        --> Write File: " + path)
            
    def _sortAlphabetically(self, w):
        return sorted(w)
    
    def _calculate(self, indicator):
        me = set()
        print("    --> Calculate Characteristic Evidence for: " + indicator)
        for action in self._actions:
            if(action.name != indicator):
                print("        --> Action for Evidence Sum: " + action.name)
                me = me | action.set

        for action in self._actions:
            if(action.name == indicator):
                action.ceSet = action.set - me

    def process(self, pathME, pathCE):

        # Go through all me files to get all actions
        for path, dirs, files in os.walk(pathME):
            for file in files:
                # Extract the action name from the file
                actionName = file.split(".")[0]
                self._actions.append(CEAction(actionName))
                p = os.path.join(pathME, actionName + ".me")
                me = open(p, "r", encoding="utf-8")
                lines = me.readlines()
                self._initSet(actionName, lines)

        # Write *.ce   
        for action in self._actions:
            if(action.name != "noise"):    
                self._calculate(action.name)
                p = os.path.join(pathCE, action.name + ".ce")
                self._writeToFile(action.name, p)


class CEAction():
    def __init__(self, name) -> None:
        self.name = name
        self.set = set()
        self.ceSet = set()

    def __repr__(self):
        return f"CEAction(Name: {self.name})"


import os

class PrepareEvidence():
    def __init__(self):
        self.currentType = ""
        self.fileType = { "NEW_FILES": "New files", "DEL_FILES": "Deleted files", 
                          "RE_FILES": "Renamed files", "MOD_FILES": "Files with modified contents",
                          "PROP_FILES": "Files with changed properties"}
        self._idiff = []    # Initial list - Contains all the lines of the idiff file
        self._list = []     # processed list - Contains the lines after processing with the given logic
        self._unique = []   # unique list - Contains only unique lines
        self._sorted = []   # sorted list - Contains alphabetically sorted lines
    
    def _checkFileType(self, line):
        if(self.fileType["NEW_FILES"] in line):
                self.currentType = self.fileType["NEW_FILES"]
        elif(self.fileType["DEL_FILES"] in line):
                self.currentType = self.fileType["DEL_FILES"]
        elif(self.fileType["RE_FILES"] in line):
                self.currentType = self.fileType["RE_FILES"]
        elif(self.fileType["MOD_FILES"] in line):
                self.currentType = self.fileType["MOD_FILES"]
        elif(self.fileType["PROP_FILES"] in line):
                self.currentType = self.fileType["PROP_FILES"]

    def _reset(self):
        self._idiff.clear()
        self._list.clear()
        self._unique.clear()
        self._sorted.clear()

    def _sortAlphabetically(self):
        self._sorted = sorted(self._unique)

    def _makeUnique(self):
        self._unique = list(set(self._list))

    def _addLine(self, line):
        self._list.append(line)

    def _writeToFile(self, path):
        file = open(path, "w", encoding="utf-8")
        for line in self._sorted:
            file.write(line + "\n")
        print("    --> Write File: " + path)

    def _setIdiffData(self, idiff):
        self._idiff = idiff

    def _sanitize(self):
        for line in self._idiff:
            if(line == ""):
                self._idiff.remove("")
            elif(line == "\n"):
                self._idiff.remove("\n")

    def _process(self):
        for line in self._idiff:
            # Checks the current file type and set the current file type
            self._checkFileType(line)

            # Is it "New Files"...
            if(self.currentType == self.fileType["NEW_FILES"]):
                if(line != "" and line != "\n"):
                    newFile = line.split("\t")
                    if(len(newFile) == 3):
                        for ts in ["m", "a", "c", "cr"]:
                            self._addLine(newFile[1] + "\t" + ts)

            # Is it "Deleted Files"...
            elif(self.currentType == self.fileType["DEL_FILES"]):
                if(line != "" and line != "\n"):
                    delFile = line.split("\t")
                    if(len(delFile) == 3):
                            self._addLine(delFile[1] + "\t" + "d")

            # Is it "Renamed Files", "Files with modified contents" oder
            # "Files with changed properties"...
            elif(self.currentType == self.fileType["RE_FILES"] 
                or self.currentType == self.fileType["MOD_FILES"]
                or self.currentType == self.fileType["PROP_FILES"]):
                if(line != "" and line != "\n"):
                    fi = line.split("\t")
                    if("mtime" in line):
                        self._addLine(fi[0] + "\t" + "m")
                    elif("ctime" in line):
                        self._addLine(fi[0] + "\t" + "c")
                    elif("atime" in line):
                        self._addLine(fi[0] + "\t" + "a")
                    elif("crtime" in line):
                        self._addLine(fi[0] + "\t" + "cr")

    def process(self, pathGE, pathPE):
        for path, dirs, files in os.walk(pathGE):
            for file in files:
                self._reset()
                # File ending of new file
                newFileName = file.replace("idiff", "pe")
                # Full path
                fullPath = os.path.join(path, file)
                # Opens the specific .idiff file
                idiff = open(fullPath, "r", encoding="utf-8")
                # Reads all the lines from the idiff file
                lines = idiff.readlines()
                # Sets the readed lines
                self._setIdiffData(lines)
                # Removes space and newlines
                self._sanitize() 
                # Processes lines with the given logic
                self._process()
                # Remove duplicates
                self._makeUnique()
                # Sort lines alphabetically
                self._sortAlphabetically()
                # Write to new file
                p = os.path.join(pathPE, newFileName)
                self._writeToFile(p)  


class Line():
    def __init__(self, path, timestamp):
        self.path = path
        self.timestamp = timestamp
        self.occurence = 1

    def stringify(self):
        return self.path + "\t" + self.timestamp + "\t" + str(self.occurence)
     
    def __repr__(self):
        return f"Line(Path: {self.path}, Timestamp: {self.timestamp}, Occurence: {self.occurence})"


import os
import time
import shutil

from datetime import datetime

class Controller():
    def __init__(self, output) -> None:
        self.startTime = time.time()
        self.workingDir = os.path.join(os.getcwd(), output) 
        self.pathPE = os.path.join(self.workingDir, "pe")
        self.pathME = os.path.join(self.workingDir, "me")
        self.pathCE = os.path.join(self.workingDir, "ce")

    def _createFolders(self):
        os.mkdir(self.workingDir)
        os.mkdir(self.pathPE)
        os.mkdir(self.pathME)
        os.mkdir(self.pathCE)

    def _clearFolders(self):
        if os.path.isdir(self.pathPE):
            shutil.rmtree(self.pathPE)
        if os.path.isdir(self.pathME):
            shutil.rmtree(self.pathME)
        if os.path.isdir(self.pathCE):
            shutil.rmtree(self.pathCE)
        if os.path.isdir(self.workingDir):
            shutil.rmtree(self.workingDir)

    def scaffold(self):
        self._clearFolders()
        self._createFolders()

    def printHeader(self):
        print("################################################################################")
        print("")
        print("evidence by 5f0")
        print("Differential analysis of DFXML idifference2.py output")
        print("")
        print("Current working directory: " + os.getcwd())
        print("")
        print("Datetime: " + datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
        print("")
        print("################################################################################")

    def printScaffolding(self):
        print("")
        print("--> Creating folder structure")
        print("    --> " + self.workingDir)

    def printPhase(self, phase):
        print("")
        print(phase)

    def printExecutionTime(self):
        end = time.time()
        print("")
        print("################################################################################")
        print("")
        print("Execution Time: " + str(end-self.startTime)[0:8] + " sec")
        print("")


import sys
import argparse

from evidence.Controller import Controller
from evidence.phase.PrepareEvidence import PrepareEvidence
from evidence.phase.MergeEvidence import MergeEvidence
from evidence.phase.CharacterizeEvidence import CharacterizeEvidence

def main(args_=None):
    """The main routine."""
    if args_ is None:
        args_ = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument("--path", "-p", type=str, required=True, help="Path to idifference output dir")
    parser.add_argument("--output", "-o", type=str, default="output", help="Path to result dir")
    parser.add_argument("--occurence", "-c", type=int, default=2, help="The number of occurences in *.me to be used in ce processing")
    args = parser.parse_args()

    ctrl = Controller(args.output)
    ctrl.printHeader()
    ctrl.printScaffolding()
    ctrl.scaffold()
    

    ctrl.printPhase("--> Prepare Evidence")
    pe = PrepareEvidence()
    pe.process(args.path, ctrl.pathPE)

    ctrl.printPhase("--> Merge Evidence")
    me = MergeEvidence()
    me.process(ctrl.pathPE, ctrl.pathME)

    ctrl.printPhase("--> Characterize Evidence")
    ce = CharacterizeEvidence(args.occurence)
    ce.process(ctrl.pathME, ctrl.pathCE)

    ctrl.printExecutionTime()


if __name__ == "__main__":
    sys.exit(main())



class MEAction():
    def __init__(self, name) -> None:
        self.name = name
        self.occurence = 0
        self.merged = []

    def increment(self):
        self.occurence += 1

    def __repr__(self):
        return f"MEAction(Name: {self.name}, Occurence: {str(self.occurence)})"


import os

from string import Template

from evidence.model.Line import Line
from evidence.model.MEAction import MEAction

class MergeEvidence():
    def __init__(self) -> None:
        self._actions = []
        self.nameTemplate = Template("$action.$number.pe")

    def _add(self, indicator, line):
        self._getListForIndicator(indicator).append(line)
        
    def _getListForIndicator(self, indicator):
        result = None
        for action in self._actions:
            if(indicator == action.name):
                result = action.merged
        return result
    
    def _init(self, indicator, lines):
        for line in lines:
            parts = line.split("\t")
            self._add(indicator, Line(path=parts[0], timestamp=parts[1].replace("\n", "")))
    
    def _compare(self, indicator, lines):
            temp = []
            # for each line in the file to be compared
            for line in lines:
                # Split into parts
                parts = line.split("\t")
                # Flag indicating if a line is already present
                isMatched = False
                # For each entry in a
                for line in self._getListForIndicator(indicator):
                    # Check if the line is already present
                    if(line.path == parts[0] and line.timestamp == parts[1].replace("\n", "")):
                        # if yes, increment occurense
                        line.occurence = line.occurence + 1
                        # set flag that hit occured
                        isMatched = True
                        break
                # If there was no hit, line need to be added 
                if(isMatched == False):
                    temp.append(Line(path=parts[0], timestamp=parts[1].replace("\n", "")))
            # Add the additional lines
            for n in temp:
                self._add(indicator, n)
            # Clears the additional lines list
            temp.clear()  

    def _writeToFile(self, indicator, path):
        w = []
        file = open(path, "w", encoding="utf-8")
        for line in self._getListForIndicator(indicator):
            w.append(line.stringify() + "\n")
  
        for line in self._sortAlphabetically(w):
            file.write(line)

        print("    --> Write File: " + path)
        
    def _sortAlphabetically(self, w):
        return sorted(w)

    def process(self, pathPE, pathME): 
        # Go through all pe files to get all actions
        for path, dirs, files in os.walk(pathPE):
            for file in files:
                # Extract the action name from the file
                actionName = file.split(".")[0]
                # If there is no action yet
                if(len(self._actions) == 0):
                    a = MEAction(actionName)
                    a.increment()
                    self._actions.append(a)
                else:
                    # If there are actions...
                    actionListed = False
                    for action in self._actions:
                        # Check if action is already created
                        if(action.name == actionName):
                            action.increment()
                            actionListed = True
                            break
                    # If action is not created yet...
                    if(not actionListed):
                        a = MEAction(actionName)
                        a.increment()
                        self._actions.append(a)

        # Read *.1.pe
        for action in self._actions:
            if(action.name != "noise"):
                name = self.nameTemplate.substitute(action=action.name, number="1")
                p = os.path.join(pathPE, name)
                pe = open(p, "r", encoding="utf-8")
                lines = pe.readlines()
                self._init(action.name, lines)

        for action in self._actions:
            if(action.occurence == 1):
                # Check if action only have one occurence.
                # If so, comparison is not needed
                p_ = os.path.join(pathME, action.name + ".me")
                self._writeToFile(action.name, p_)
            else:
                # Read *.*.pe, compare to *.1.pe and write *.me
                if(action.name != "noise"):
                    for i in range(2, (action.occurence+1)):
                        name = self.nameTemplate.substitute(action=action.name, number=i)
                        p = os.path.join(pathPE, name)
                        pe = open(p, "r", encoding="utf-8")
                        lines = pe.readlines()
                        self._compare(action.name, lines)
                    p_ = os.path.join(pathME, action.name + ".me")
                    self._writeToFile(action.name,  p_)
                
        # Read noise.pe and write noise.me
        p = os.path.join(pathPE, "noise.1.pe")
        pe = open(p, "r", encoding="utf-8")
        lines = pe.readlines()
        self._init("noise", lines)
        p = os.path.join(pathME, "noise.me")
        self._writeToFile("noise", p)